from __future__ import annotations

from abc import abstractmethod

from ..parametrization import Parametrization

__all__ = ['Multidimensional']


class Multidimensional(Parametrization):
    @abstractmethod
    def dimensions(self) -> int:
        raise NotImplementedError



from __future__ import annotations

from typing import Any, Generic, TypeVar

import jax
import jax.numpy as jnp
from jax.nn import softplus
from jax.random import KeyArray
from jax.scipy import special as jss
from tjax import JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..interfaces.multidimensional import Multidimensional
from ..interfaces.samplable import Samplable
from ..mixins.exp_to_nat import ExpToNat
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import VectorSupport, distribution_parameter

__all__: list[str] = []


EP = TypeVar('EP', bound='DirichletCommonEP[Any]')


@dataclass
class DirichletCommonNP(HasEntropyNP,
                        Samplable,
                        Multidimensional,
                        NaturalParametrization['DirichletCommonEP[Any]', JaxRealArray]):
    alpha_minus_one: JaxRealArray = distribution_parameter(VectorSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.alpha_minus_one.shape[:-1]

    @override
    def log_normalizer(self) -> JaxRealArray:
        q = self.alpha_minus_one
        return (jnp.sum(jss.gammaln(q + 1.0), axis=-1)
                - jss.gammaln(jnp.sum(q, axis=-1) + self.dimensions()))

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        return jax.random.dirichlet(key, 1.0 + self.alpha_minus_one, shape)[..., :-1]

    @override
    def dimensions(self) -> int:
        return self.alpha_minus_one.shape[-1]

    def _exp_helper(self) -> JaxRealArray:
        q = self.alpha_minus_one
        return jss.digamma(q + 1.0) - jss.digamma(jnp.sum(q, axis=-1, keepdims=True) + q.shape[-1])


NP = TypeVar('NP', bound=DirichletCommonNP)


@dataclass
class DirichletCommonEP(HasEntropyEP[NP],
                        ExpToNat[NP, JaxRealArray],
                        Multidimensional, Generic[NP]):
    mean_log_probability: JaxRealArray = distribution_parameter(VectorSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_log_probability.shape[:-1]

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        return jnp.zeros(self.shape)

    @override
    def initial_search_parameters(self) -> JaxRealArray:
        return jnp.zeros(self.mean_log_probability.shape)

    @override
    def search_gradient(self, search_parameters: JaxRealArray) -> JaxRealArray:
        return self._natural_gradient(self.search_to_natural(search_parameters)).alpha_minus_one

    @override
    def dimensions(self) -> int:
        return self.mean_log_probability.shape[-1]

    @classmethod
    def _transform_nat_helper(cls, search_parameters: JaxRealArray) -> JaxRealArray:
        # Run Newton's method on the whole real hyperspace.
        return softplus(search_parameters) - 1.0



from ._src import (BooleanRing, ComplexField, ExpectationParametrization, IntegralRing,
                   NaturalParametrization, RealField, Ring, ScalarSupport, SquareMatrixSupport,
                   Support, SymmetricMatrixSupport, VectorSupport)
from ._src.distributions import (BernoulliEP, BernoulliNP, BetaEP, BetaNP, ChiEP, ChiNP,
                                 ChiSquareEP, ChiSquareNP, ComplexCircularlySymmetricNormalEP,
                                 ComplexCircularlySymmetricNormalNP,
                                 ComplexMultivariateUnitNormalEP, ComplexMultivariateUnitNormalNP,
                                 ComplexNormalEP, ComplexNormalNP, ComplexUnitNormalEP,
                                 ComplexUnitNormalNP, DirichletEP, DirichletNP, ExponentialEP,
                                 ExponentialNP, GammaEP, GammaNP, GammaVP, GeneralizedDirichletEP,
                                 GeneralizedDirichletNP, GeometricEP, GeometricNP,
                                 IsotropicNormalEP, IsotropicNormalNP, LogarithmicEP, LogarithmicNP,
                                 MultinomialEP, MultinomialNP, MultivariateDiagonalNormalEP,
                                 MultivariateDiagonalNormalNP, MultivariateDiagonalNormalVP,
                                 MultivariateFixedVarianceNormalEP,
                                 MultivariateFixedVarianceNormalNP, MultivariateNormalEP,
                                 MultivariateNormalNP, MultivariateNormalVP,
                                 MultivariateUnitNormalEP, MultivariateUnitNormalNP,
                                 NegativeBinomialEP, NegativeBinomialNP, NormalEP, NormalNP,
                                 NormalVP, PoissonEP, PoissonNP, RayleighEP, RayleighNP,
                                 UnitNormalEP, UnitNormalNP, VonMisesFisherEP, VonMisesFisherNP,
                                 WeibullEP, WeibullNP)
from ._src.interfaces import (HasConjugatePrior, HasGeneralizedConjugatePrior, Multidimensional,
                              Samplable)
from ._src.mixins import HasEntropyEP, HasEntropyNP
from ._src.scipy_replacement import (ScipyComplexMultivariateNormal, ScipyComplexNormal,
                                     ScipyDirichlet, ScipyGeneralizedDirichlet,
                                     ScipyMultivariateNormal, ScipyVonMises)

__all__ = ['ExpectationParametrization', 'HasConjugatePrior', 'HasGeneralizedConjugatePrior',
           'Multidimensional', 'NaturalParametrization', 'Samplable', 'ScalarSupport',
           'SquareMatrixSupport', 'Support', 'SymmetricMatrixSupport', 'VectorSupport',
           'BernoulliEP', 'BernoulliNP', 'BetaEP', 'BetaNP', 'ChiEP', 'ChiNP', 'ChiSquareEP',
           'ChiSquareNP', 'ComplexCircularlySymmetricNormalEP',
           'ComplexCircularlySymmetricNormalNP', 'ComplexMultivariateUnitNormalEP',
           'ComplexMultivariateUnitNormalNP', 'ComplexUnitNormalEP', 'ComplexUnitNormalNP',
           'ComplexNormalEP', 'ComplexNormalNP', 'DirichletEP', 'DirichletNP', 'ExponentialEP',
           'ExponentialNP', 'GammaEP', 'GammaNP', 'GammaVP', 'GeneralizedDirichletEP',
           'GeneralizedDirichletNP', 'GeometricEP', 'GeometricNP', 'IsotropicNormalEP',
           'IsotropicNormalNP', 'LogarithmicEP', 'LogarithmicNP', 'MultinomialEP', 'MultinomialNP',
           'MultivariateDiagonalNormalEP', 'MultivariateDiagonalNormalNP',
           'MultivariateDiagonalNormalVP', 'MultivariateFixedVarianceNormalEP',
           'MultivariateFixedVarianceNormalNP', 'MultivariateNormalEP', 'MultivariateNormalNP',
           'MultivariateNormalVP', 'NormalVP', 'MultivariateUnitNormalEP',
           'MultivariateUnitNormalNP', 'NegativeBinomialEP', 'NegativeBinomialNP', 'NormalEP',
           'NormalNP', 'PoissonEP', 'PoissonNP', 'RayleighEP', 'RayleighNP', 'VonMisesFisherEP',
           'VonMisesFisherNP', 'WeibullEP', 'WeibullNP', 'ScipyComplexMultivariateNormal',
           'ScipyComplexNormal', 'ScipyDirichlet', 'ScipyGeneralizedDirichlet',
           'ScipyMultivariateNormal', 'ScipyVonMises', 'HasEntropyNP', 'HasEntropyEP',
           'UnitNormalNP', 'UnitNormalEP', 'Ring', 'IntegralRing', 'BooleanRing', 'RealField',
           'ComplexField']



from __future__ import annotations

from typing import Any

import jax
import jax.numpy as jnp
import numpy as np
from jax.random import KeyArray
from tjax import JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import ScalarSupport, distribution_parameter

__all__ = ['NormalNP', 'NormalEP', 'NormalVP']


@dataclass
class NormalNP(HasEntropyNP, NaturalParametrization['NormalEP', JaxRealArray]):
    """The natural parametrization of the normal distribution.

    Args:
        mean_times_precision: The mean times the precision.
        negative_half_precision: -0.5 times the precision.
    """
    mean_times_precision: JaxRealArray = distribution_parameter(ScalarSupport())
    negative_half_precision: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_times_precision.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        return (-jnp.square(self.mean_times_precision) / (4.0 * self.negative_half_precision)
                + 0.5 * jnp.log(-np.pi / self.negative_half_precision))

    @override
    def to_exp(self) -> NormalEP:
        mean = -self.mean_times_precision / (2.0 * self.negative_half_precision)
        second_moment = jnp.square(mean) - 0.5 / self.negative_half_precision
        return NormalEP(mean, second_moment)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        return jnp.zeros(x.shape)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> NormalEP:
        return NormalEP(x, jnp.square(x))

    def to_var(self) -> NormalVP:
        return self.to_exp().to_var()


@dataclass
class NormalEP(HasEntropyEP[NormalNP],
               ExpectationParametrization[NormalNP],
               Samplable):
    """The expectation parametrization of the normal distribution.

    Args:
        mean: The mean.
        second_moment: The second moment.
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport())
    second_moment: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[NormalNP]:
        return NormalNP

    @override
    def to_nat(self) -> NormalNP:
        return NormalNP(self.mean / self.variance(), -0.5 / self.variance())

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        return jnp.zeros(self.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        deviation = jnp.sqrt(self.variance())
        return jax.random.normal(key, shape) * deviation + self.mean

    def variance(self) -> JaxRealArray:
        return self.second_moment - jnp.square(self.mean)

    def to_var(self) -> NormalVP:
        return NormalVP(self.mean, self.variance())


@dataclass
class NormalVP(Samplable):
    """The variance parametrization of the normal distribution.

    Args:
        mean: The mean.
        variance: The variance.
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport())
    variance: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.mean.shape
        else:
            shape = self.mean.shape
        deviation = jnp.sqrt(self.variance)
        return jax.random.normal(key, shape) * deviation + self.mean

    def log_pdf(self, x: JaxRealArray) -> JaxRealArray:
        return self.to_nat().log_pdf(x)

    def pdf(self, x: JaxRealArray) -> JaxRealArray:
        return self.to_nat().pdf(x)

    def to_exp(self) -> NormalEP:
        second_moment = self.variance + jnp.square(self.mean)
        return NormalEP(self.mean, second_moment)

    def to_nat(self) -> NormalNP:
        precision = 1.0 / self.variance
        mean_times_precision = self.mean * precision
        negative_half_precision = -0.5 * precision
        return NormalNP(mean_times_precision, negative_half_precision)



from __future__ import annotations

from typing import Any

import jax.numpy as jnp
from jax.nn import softplus
from tjax import JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..mixins.exp_to_nat import ExpToNat
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter

__all__ = ['LogarithmicNP', 'LogarithmicEP']


log_probability_floor = -50.0
log_probability_ceiling = -1e-7


@dataclass
class LogarithmicNP(NaturalParametrization['LogarithmicEP', JaxRealArray]):
    """The natural parametrization of the logarithmic distribution.

    Args:
        log_probability: log(p).
    """
    log_probability: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.log_probability.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        return jnp.log(-jnp.log1p(-jnp.exp(self.log_probability)))

    @override
    def to_exp(self) -> LogarithmicEP:
        probability = jnp.exp(self.log_probability)
        chi = jnp.where(self.log_probability < log_probability_floor, 1.0,
                        jnp.where(self.log_probability > log_probability_ceiling, jnp.inf,
                                  probability / (jnp.expm1(self.log_probability)
                                                 * jnp.log1p(-probability))))
        return LogarithmicEP(chi)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        return -jnp.log(x)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> LogarithmicEP:
        return LogarithmicEP(x)


@dataclass
class LogarithmicEP(ExpToNat[LogarithmicNP, JaxRealArray],
                    ExpectationParametrization[LogarithmicNP]):
    """The expectation parametrization of the logarithmic distribution.

    Args:
        chi: -(p / (1-p)) * log(1-p).
    """
    chi: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.chi.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[LogarithmicNP]:
        return LogarithmicNP

    # The expected_carrier_measure is unknown.

    @override
    def initial_search_parameters(self) -> JaxRealArray:
        return jnp.zeros(self.chi.shape)

    @override
    def search_to_natural(self, search_parameters: JaxRealArray) -> LogarithmicNP:
        # Run Newton's method on the whole real line.
        return LogarithmicNP(-softplus(-search_parameters))

    @override
    def search_gradient(self, search_parameters: JaxRealArray) -> JaxRealArray:
        return self._natural_gradient(self.search_to_natural(search_parameters)).log_probability

    @override
    def to_nat(self) -> LogarithmicNP:
        z: LogarithmicNP = super().to_nat()
        return LogarithmicNP(jnp.where(self.chi < 1.0,  # noqa: PLR2004
                                       jnp.nan,
                                       jnp.where(self.chi == 1.0,  # noqa: PLR2004
                                                 jnp.inf,
                                                 z.log_probability)))



from __future__ import annotations

from typing import Any

import jax
import jax.numpy as jnp
import numpy as np
from jax.random import KeyArray
from tjax import JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.multidimensional import Multidimensional
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import VectorSupport, distribution_parameter

__all__ = ['MultivariateDiagonalNormalNP', 'MultivariateDiagonalNormalEP',
           'MultivariateDiagonalNormalVP']


@dataclass
class MultivariateDiagonalNormalNP(HasEntropyNP,
                                   NaturalParametrization['MultivariateDiagonalNormalEP',
                                                          JaxRealArray],
                                   Multidimensional):
    """The natural parametrization of the normal distribution with diagonal variance.

    Args:
        mean_times_precision: E(x) / Var(x).
        negative_half_precision: The diagonal elements of -0.5 / Var(x).
    """
    mean_times_precision: JaxRealArray = distribution_parameter(VectorSupport())
    negative_half_precision: JaxRealArray = distribution_parameter(VectorSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_times_precision.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        components = (-jnp.square(self.mean_times_precision) / (4.0 * self.negative_half_precision)
                      + 0.5 * jnp.log(-np.pi / self.negative_half_precision))
        return jnp.sum(components, axis=-1)

    @override
    def to_exp(self) -> MultivariateDiagonalNormalEP:
        mean = -self.mean_times_precision / (2.0 * self.negative_half_precision)
        second_moment = jnp.square(mean) - 0.5 / self.negative_half_precision
        return MultivariateDiagonalNormalEP(mean, second_moment)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        return jnp.zeros(x.shape[:-1])

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> MultivariateDiagonalNormalEP:
        return MultivariateDiagonalNormalEP(x, jnp.square(x))

    @override
    def dimensions(self) -> int:
        return self.mean_times_precision.shape[-1]


@dataclass
class MultivariateDiagonalNormalEP(HasEntropyEP[MultivariateDiagonalNormalNP],
                                   ExpectationParametrization[MultivariateDiagonalNormalNP],
                                   Multidimensional,
                                   Samplable):
    """The expectation parametrization of the normal distribution with diagonal variance.

    Args:
        mean: E(x).
        second moment: The diagonal elements of E(x^2).
    """
    mean: JaxRealArray = distribution_parameter(VectorSupport())
    second_moment: JaxRealArray = distribution_parameter(VectorSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[MultivariateDiagonalNormalNP]:
        return MultivariateDiagonalNormalNP

    @override
    def to_nat(self) -> MultivariateDiagonalNormalNP:
        return MultivariateDiagonalNormalNP(self.mean / self.variance(), -0.5 / self.variance())

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        return jnp.zeros(self.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_variance_parametrization().sample(key, shape)

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def variance(self) -> JaxRealArray:
        return self.second_moment - jnp.square(self.mean)

    def to_variance_parametrization(self) -> MultivariateDiagonalNormalVP:
        return MultivariateDiagonalNormalVP(self.mean, self.variance())


@dataclass
class MultivariateDiagonalNormalVP(Samplable, Multidimensional):
    """The variance parametrization of the normal distribution with diagonal variance.

    Args:
        mean: E(x).
        variance: The diagonal elements of Var(x).
    """
    mean: JaxRealArray = distribution_parameter(VectorSupport())
    variance: JaxRealArray = distribution_parameter(VectorSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.mean.shape
        else:
            shape = self.mean.shape
        deviation = jnp.sqrt(self.variance)
        return jax.random.normal(key, shape) * deviation + self.mean

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def pdf(self, x: JaxRealArray) -> JaxRealArray:
        return self.to_nat().pdf(x)

    def to_exp(self) -> MultivariateDiagonalNormalEP:
        second_moment = self.variance + jnp.square(self.mean)
        return MultivariateDiagonalNormalEP(self.mean, second_moment)

    def to_nat(self) -> MultivariateDiagonalNormalNP:
        precision = 1.0 / self.variance
        mean_times_precision = self.mean * precision
        negative_half_precision = -0.5 * precision
        return MultivariateDiagonalNormalNP(mean_times_precision, negative_half_precision)



from __future__ import annotations

import math
from typing import Any

import jax
import jax.numpy as jnp
from jax.random import KeyArray
from tjax import JaxComplexArray, JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.multidimensional import Multidimensional
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import (SymmetricMatrixSupport, VectorSupport, complex_field,
                          distribution_parameter)

__all__ = ['ComplexCircularlySymmetricNormalNP', 'ComplexCircularlySymmetricNormalEP']


def _broadcasted_outer_c(x: JaxComplexArray) -> JaxComplexArray:
    return jnp.einsum("...i,...j->...ij", x, x.conjugate())


@dataclass
class ComplexCircularlySymmetricNormalNP(
        HasEntropyNP,
        NaturalParametrization['ComplexCircularlySymmetricNormalEP', JaxComplexArray],
        Multidimensional,
        Samplable):
    """The natural parameters of the circularly symmetric complex multivariate normal distribution.

    This has zero mean and and zero pseudo-variance. This is a curved exponential family.

    Args:
        negative_precision: -1/Var(x).
    """
    negative_precision: JaxComplexArray = distribution_parameter(
        SymmetricMatrixSupport(hermitian=True))
    # S = -1/negative_precision, U = 0
    # P = S.conjugate, R = 0
    # H = -1/S, J = 0
    # K = 0, L = S/2
    # eta = 0
    # Leta = 0

    @property
    @override
    def shape(self) -> Shape:
        return self.negative_precision.shape[:-2]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport(ring=complex_field)

    @override
    def log_normalizer(self) -> JaxRealArray:
        log_det_s = jnp.log(jnp.linalg.det(-self.negative_precision).real)
        return -log_det_s + self.dimensions() * math.log(math.pi)

    @override
    def to_exp(self) -> ComplexCircularlySymmetricNormalEP:
        return ComplexCircularlySymmetricNormalEP(
            jnp.linalg.inv(-self.negative_precision).conjugate())

    @override
    def carrier_measure(self, x: JaxComplexArray) -> JaxRealArray:
        return jnp.zeros(x.shape[:-1])

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxComplexArray, **fixed_parameters: Any
                              ) -> ComplexCircularlySymmetricNormalEP:
        return ComplexCircularlySymmetricNormalEP(_broadcasted_outer_c(x))

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxComplexArray:
        return self.to_exp().sample(key, shape)

    @override
    def dimensions(self) -> int:
        return self.negative_precision.shape[-1]


@dataclass
class ComplexCircularlySymmetricNormalEP(
        HasEntropyEP[ComplexCircularlySymmetricNormalNP],
        ExpectationParametrization[ComplexCircularlySymmetricNormalNP],
        Multidimensional,
        Samplable):
    """The expectation parameters of the circularly symmetric complex normal distribution.

    This has zero mean and and zero pseudo-variance. This is a curved exponential family.

    Args:
        variance: Var(x).
    """
    variance: JaxComplexArray = distribution_parameter(SymmetricMatrixSupport(hermitian=True))

    @property
    @override
    def shape(self) -> Shape:
        return self.variance.shape[:-2]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport(ring=complex_field)

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[ComplexCircularlySymmetricNormalNP]:
        return ComplexCircularlySymmetricNormalNP

    @override
    def to_nat(self) -> ComplexCircularlySymmetricNormalNP:
        return ComplexCircularlySymmetricNormalNP(
            -jnp.linalg.inv(self.variance).conjugate())

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        return jnp.zeros(self.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxComplexArray:
        n = self.dimensions()
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        xy_rvs = jax.random.multivariate_normal(key,
                                                self._multivariate_normal_mean(),
                                                self._multivariate_normal_cov(),
                                                shape)
        return xy_rvs[..., :n] + 1j * xy_rvs[..., n:]

    @override
    def dimensions(self) -> int:
        return self.variance.shape[-1]

    def _multivariate_normal_mean(self) -> JaxRealArray:
        """Return the mean of a corresponding real distribution with double the size."""
        n = self.dimensions()
        return jnp.zeros((*self.shape, n * 2))

    def _multivariate_normal_cov(self) -> JaxRealArray:
        """Return the covariance of a corresponding real distribution with double the size."""
        gamma_r = 0.5 * self.variance.real
        gamma_i = 0.5 * self.variance.imag
        return jnp.block([[gamma_r, -gamma_i], [gamma_i, gamma_r]])



from __future__ import annotations

import numpy as np
import scipy.stats as ss
from tjax import NumpyRealArray
from typing_extensions import override

from .shaped_distribution import ShapedDistribution

__all__ = ['ScipyVonMises']


class ScipyVonMises(ShapedDistribution[ss.vonmises]):  # pyright: ignore
    """This class allows distributions having a non-empty shape."""
    @override
    def __init__(self, kappa: NumpyRealArray, loc: NumpyRealArray | None = None):
        if loc is None:
            loc = np.zeros_like(kappa)
        shape = np.broadcast(kappa, loc).shape
        rvs_shape = ()
        dtype = np.result_type(kappa.dtype, loc.dtype)
        kappa = np.broadcast_to(kappa, shape).astype(dtype)
        loc = np.broadcast_to(loc, shape).astype(dtype)
        objects = np.empty(shape, dtype=np.object_)
        for i in np.ndindex(*shape):
            objects[i] = ss.vonmises(kappa[i], loc[i])
        super().__init__(shape, rvs_shape, dtype, objects)



from __future__ import annotations

from typing import Any

import jax
import jax.numpy as jnp
import numpy as np
from jax.random import KeyArray
from jax.scipy import special as jss
from tjax import JaxBooleanArray, JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.conjugate_prior import HasConjugatePrior
from ..interfaces.samplable import Samplable
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, boolean_ring, distribution_parameter
from .beta import BetaNP

__all__ = ['BernoulliNP', 'BernoulliEP']


@dataclass
class BernoulliNP(HasEntropyNP,
                  NaturalParametrization['BernoulliEP', JaxRealArray]):
    """The natural parametrization of the Bernoulli distribution.

    Args:
        log_odds: log(p / (1-p)).
    """
    log_odds: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.log_odds.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=boolean_ring)

    @override
    def log_normalizer(self) -> JaxRealArray:
        return jnp.logaddexp(self.log_odds, 0.0)

    @override
    def to_exp(self) -> BernoulliEP:
        return BernoulliEP(jss.expit(self.log_odds))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        return jnp.zeros(x.shape)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> BernoulliEP:
        return BernoulliEP(x)

    def nat_to_probability(self) -> JaxRealArray:
        p = jss.expit(self.log_odds)
        final_p = 1.0 - p
        return jnp.stack([p, final_p], axis=-1)

    def nat_to_surprisal(self) -> JaxRealArray:
        total_p = self.nat_to_probability()
        return -jnp.log(total_p)


@dataclass
class BernoulliEP(HasEntropyEP[BernoulliNP],
                  HasConjugatePrior,
                  Samplable,
                  ExpectationParametrization[BernoulliNP]):
    """The expectation parametrization of the Bernoulli distribution.

    Args:
        probability: p.
    """
    probability: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.probability.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=boolean_ring)

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[BernoulliNP]:
        return BernoulliNP

    @override
    def to_nat(self) -> BernoulliNP:
        return BernoulliNP(jss.logit(self.probability))

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        return jnp.zeros(self.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxBooleanArray:
        if shape is not None:
            shape += self.shape
        return jax.random.bernoulli(key, self.probability, shape)

    @override
    def conjugate_prior_distribution(self, n: JaxRealArray) -> BetaNP:
        reshaped_n = n[..., np.newaxis]
        return BetaNP(reshaped_n * jnp.stack([self.probability, (1.0 - self.probability)], axis=-1))

    @override
    def conjugate_prior_observation(self) -> JaxRealArray:
        return self.probability



from __future__ import annotations

from typing import Any

import jax
import jax.numpy as jnp
from jax.nn import softplus
from jax.random import KeyArray
from jax.scipy import special as jss
from tjax import JaxRealArray, Shape, inverse_softplus
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.samplable import Samplable
from ..mixins.exp_to_nat import ExpToNat
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter
from ..parametrization import Parametrization

__all__ = ['GammaNP', 'GammaEP', 'GammaVP']


@dataclass
class GammaNP(HasEntropyNP,
              Samplable,
              NaturalParametrization['GammaEP', JaxRealArray]):
    """The natural parametrization of the Gamma distribution.

    Args:
        negative_rate: The negative rate.
        shape_minus_one: The shape minus one.
    """
    negative_rate: JaxRealArray = distribution_parameter(ScalarSupport())
    shape_minus_one: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.negative_rate.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        shape = self.shape_minus_one + 1.0
        return jss.gammaln(shape) - shape * jnp.log(-self.negative_rate)

    @override
    def to_exp(self) -> GammaEP:
        shape = self.shape_minus_one + 1.0
        return GammaEP(-shape / self.negative_rate,
                       jss.digamma(shape) - jnp.log(-self.negative_rate))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        return jnp.zeros(x.shape)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> GammaEP:
        return GammaEP(x, jnp.log(x))

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        return -jax.random.gamma(key, self.shape_minus_one + 1.0, shape) / self.negative_rate

    def to_var(self) -> GammaVP:
        rate = -self.negative_rate
        shape = self.shape_minus_one + 1.0
        mean = shape / rate
        variance = mean / rate
        return GammaVP(mean, variance)


@dataclass
class GammaEP(HasEntropyEP[GammaNP],
              ExpToNat[GammaNP, JaxRealArray],
              ExpectationParametrization[GammaNP]):
    """The expectation parametrization of the Gamma distribution.

    Args:
        mean: The mean: E(x).
        mean_log: The mean of the log: E(log(x)).
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport())
    mean_log: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[GammaNP]:
        return GammaNP

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        return jnp.zeros(self.shape)

    @override
    def search_to_natural(self, search_parameters: JaxRealArray) -> GammaNP:
        shape = softplus(search_parameters)
        rate = shape / self.mean
        return GammaNP(-rate, shape - 1.0)

    @override
    def search_gradient(self, search_parameters: JaxRealArray) -> JaxRealArray:
        shape = softplus(search_parameters)
        log_mean_minus_mean_log = jnp.log(self.mean) - self.mean_log
        return log_mean_minus_mean_log - jnp.log(shape) + jss.digamma(shape)
        # gradient is 1.0 / shape - jss.polygamma(1, shape)
        # where polygamma(1) is trigamma

    @override
    def initial_search_parameters(self) -> JaxRealArray:
        log_mean_minus_mean_log = jnp.log(self.mean) - self.mean_log
        initial_shape = (
            (3.0 - log_mean_minus_mean_log
             + jnp.sqrt((log_mean_minus_mean_log - 3.0) ** 2 + 24.0 * log_mean_minus_mean_log))
            / (12.0 * log_mean_minus_mean_log))
        return inverse_softplus(initial_shape)


@dataclass
class GammaVP(Parametrization):
    """The variance parametrization of the Gamma distribution.

    Args:
        mean: The mean.
        variance: The variance.
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport())
    variance: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    def to_nat(self) -> GammaNP:
        # mean = s/r
        # var = s/r^2
        rate = self.mean / self.variance
        shape = self.mean * rate
        return GammaNP(-rate, shape - 1.0)

    def to_exp(self) -> GammaEP:
        return self.to_nat().to_exp()



from __future__ import annotations

from typing import Any

import jax.numpy as jnp
from jax.random import KeyArray, chisquare
from jax.scipy import special as jss
from tjax import JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.samplable import Samplable
from ..mixins.exp_to_nat import ExpToNat
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter

__all__ = ['ChiSquareNP', 'ChiSquareEP']


@dataclass
class ChiSquareNP(HasEntropyNP,
                  Samplable,
                  NaturalParametrization['ChiSquareEP', JaxRealArray]):
    """The natural parameters of the chi-square distribution with k degrees of freedom.

    This is the gamma distribution with shape k/2 and rate 1/2.

    Args:
        k_over_two_minus_one: k/2 - 1.
    """
    k_over_two_minus_one: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.k_over_two_minus_one.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        k_over_two = self.k_over_two_minus_one + 1.0
        return jss.gammaln(k_over_two) - k_over_two * jnp.log(0.5)

    @override
    def to_exp(self) -> ChiSquareEP:
        k_over_two = self.k_over_two_minus_one + 1.0
        return ChiSquareEP(jss.digamma(k_over_two) - jnp.log(0.5))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        return -x * 0.5

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> ChiSquareEP:
        return ChiSquareEP(jnp.log(x))

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        degrees_of_freedom = (self.k_over_two_minus_one + 1.0) * 2.0
        return chisquare(key, degrees_of_freedom, shape)


# The ExpToNat mixin can be circumvented if the inverse of the digamma function were added to JAX.
@dataclass
class ChiSquareEP(HasEntropyEP[ChiSquareNP],
                  ExpToNat[ChiSquareNP, ChiSquareNP],
                  ExpectationParametrization[ChiSquareNP]):
    """The expectation parameters of the chi-square distribution with k degrees of freedom.

    This is the gamma distribution with shape k/2 and rate 1/2.

    Args:
        mean_log: E(log(x)).
    """
    mean_log: JaxRealArray = distribution_parameter(ScalarSupport())

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[ChiSquareNP]:
        return ChiSquareNP

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_log.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        q = self.to_nat()
        k_over_two = q.k_over_two_minus_one + 1.0
        return -k_over_two

    @override
    def initial_search_parameters(self) -> ChiSquareNP:
        return ChiSquareNP(jnp.zeros(self.mean_log.shape))

    @override
    def search_to_natural(self, search_parameters: ChiSquareNP) -> ChiSquareNP:
        return search_parameters

    @override
    def search_gradient(self, search_parameters: ChiSquareNP) -> ChiSquareNP:
        return self._natural_gradient(search_parameters)



from __future__ import annotations

from typing import Any

import jax
import jax.numpy as jnp
from jax.random import KeyArray
from jax.scipy import special as jss
from tjax import JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.conjugate_prior import HasConjugatePrior
from ..interfaces.samplable import Samplable
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter, integral_ring
from .gamma import GammaNP

__all__ = ['PoissonNP', 'PoissonEP']


@dataclass
class PoissonNP(NaturalParametrization['PoissonEP', JaxRealArray]):
    """The natural parametrization of the Poisson distribution.

    Args:
        log_mean: log(E(x)).
    """
    log_mean: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.log_mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=integral_ring)

    @override
    def log_normalizer(self) -> JaxRealArray:
        return jnp.exp(self.log_mean)

    @override
    def to_exp(self) -> PoissonEP:
        return PoissonEP(jnp.exp(self.log_mean))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        return -jss.gammaln(x + 1)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> PoissonEP:
        return PoissonEP(x)


@dataclass
class PoissonEP(HasConjugatePrior,
                Samplable,
                ExpectationParametrization[PoissonNP]):
    """The expectation parametrization of the Poisson distribution.

    Args:
        mean: E(x).
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=integral_ring)

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[PoissonNP]:
        return PoissonNP

    @override
    def to_nat(self) -> PoissonNP:
        return PoissonNP(jnp.log(self.mean))

    # The expected_carrier_measure is -exp(-mean) * sum over k from zero to infinity of
    #   lambda ** k * log(k!) / k! = lambda ** k * log Gamma(k+1) / Gamma(k+1)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        return jax.random.poisson(key, self.mean, shape)

    @override
    def conjugate_prior_distribution(self, n: JaxRealArray) -> GammaNP:
        return GammaNP(-n, n * self.mean)

    @override
    def conjugate_prior_observation(self) -> JaxRealArray:
        return self.mean



from __future__ import annotations

import math
from typing import Any

import jax.numpy as jnp
from tjax import JaxComplexArray, JaxRealArray, Shape, abs_square
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import ScalarSupport, complex_field, distribution_parameter

__all__ = ['ComplexNormalNP', 'ComplexNormalEP']


@dataclass
class ComplexNormalNP(HasEntropyNP,
                      NaturalParametrization['ComplexNormalEP', JaxComplexArray]):
    """The natural parametrization of the complex normal distribution.

    Args:
        mean_times_precision: E(x) / Var(x).
        precision: 1 / Var(x).
        pseudo_precision: 1 / E(x^2 - E(x)^2).
    """
    mean_times_precision: JaxComplexArray = distribution_parameter(
        ScalarSupport(ring=complex_field))
    precision: JaxRealArray = distribution_parameter(ScalarSupport())
    pseudo_precision: JaxComplexArray = distribution_parameter(ScalarSupport(ring=complex_field))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_times_precision.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=complex_field)

    @override
    def log_normalizer(self) -> JaxRealArray:
        _, s, mu = self._r_s_mu()
        det_s = s
        det_h = -self.precision
        return (-abs_square(mu) * self.precision
                - (jnp.square(mu) * self.pseudo_precision).real
                + 0.5 * jnp.log(det_s)
                - 0.5 * jnp.log(det_h)
                + math.log(math.pi))

    @override
    def to_exp(self) -> ComplexNormalEP:
        r, s, mu = self._r_s_mu()
        u = (r * s).conjugate()
        return ComplexNormalEP(mu, s + abs_square(mu), u + jnp.square(mu))

    @override
    def carrier_measure(self, x: JaxComplexArray) -> JaxRealArray:
        return jnp.zeros(x.shape)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxComplexArray, **fixed_parameters: Any
                              ) -> ComplexNormalEP:
        return ComplexNormalEP(x, abs_square(x), jnp.square(x))

    def _r_s_mu(self) -> tuple[JaxComplexArray, JaxRealArray, JaxComplexArray]:
        r = -self.pseudo_precision / self.precision
        s = 1.0 / ((abs_square(r) - 1.0) * self.precision)
        k = self.pseudo_precision / self.precision
        l_eta = 0.5 * self.mean_times_precision / ((abs_square(k) - 1.0) * self.precision)
        mu = l_eta.conjugate() - (self.pseudo_precision / self.precision).conjugate() * l_eta
        return r, s, mu


@dataclass
class ComplexNormalEP(HasEntropyEP[ComplexNormalNP],
                      ExpectationParametrization[ComplexNormalNP]):
    """The expectation parametrization of the complex normal distribution.

    Args:
        mean: E(x).
        second_moment: E(x * conjugate(x)).
        pseudo_second_moment: E(x^2).
    """
    mean: JaxComplexArray = distribution_parameter(ScalarSupport(ring=complex_field))
    second_moment: JaxRealArray = distribution_parameter(ScalarSupport())
    pseudo_second_moment: JaxComplexArray = distribution_parameter(
        ScalarSupport(ring=complex_field))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=complex_field)

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[ComplexNormalNP]:
        return ComplexNormalNP

    @override
    def to_nat(self) -> ComplexNormalNP:
        variance = self.second_moment - abs_square(self.mean)
        pseudo_variance = self.pseudo_second_moment - jnp.square(self.mean)

        r = pseudo_variance.conjugate() / variance
        p_c = variance - (r * pseudo_variance).conjugate()
        p_inv_c = 1.0 / p_c
        precision = -p_inv_c.real
        pseudo_precision = r * p_inv_c
        mean_times_precision = -2.0 * (precision * self.mean.conjugate()
                                       + pseudo_precision * self.mean)
        return ComplexNormalNP(mean_times_precision, precision, pseudo_precision)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        return jnp.zeros(self.shape)



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Iterable
from dataclasses import fields
from functools import partial, reduce
from itertools import count
from typing import Any

import jax.numpy as jnp
from tjax import JaxAbstractClass, JaxComplexArray, JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import Self

from .parameter import Support

__all__ = ['Parametrization']


@dataclass
class Parametrization(JaxAbstractClass):
    def __getitem__(self, key: Any) -> Self:
        parameters = {field.name: getattr(self, field.name)[key]
                      for field in fields(self)}
        return type(self)(**parameters)

    def flattened(self) -> JaxRealArray:
        return reduce(partial(jnp.append, axis=-1),
                      (support.flattened(value)
                       for _, value, support in self.parameters_name_value_support()))

    @classmethod
    def unflattened(cls, flattened: JaxRealArray, **fixed_parameters: Any) -> Self:
        # Solve for dimensions.
        def total_elements(dimensions: int) -> int:
            return sum(support.num_elements(dimensions)
                       for _, support in cls.parameters_name_support())

        target = flattened.shape[-1]
        dimensions = 0
        for dimensions in count():
            te = total_elements(dimensions)
            if te == target:
                break
            if te > target:
                raise ValueError

        # Unflatten.
        kwargs = fixed_parameters
        consumed = 0
        for name, support in cls.parameters_name_support():
            k = support.num_elements(dimensions)
            kwargs[name] = support.unflattened(flattened[..., consumed: consumed + k], dimensions)
            consumed += k

        return cls(**kwargs)

    def fixed_parameters(self) -> dict[str, Any]:
        return {field.name: getattr(self, field.name)
            for field in fields(self)
            if field.metadata['fixed']}

    def parameters_value_support(self) -> Iterable[tuple[JaxComplexArray, Support]]:
        """The value and support of each variable parameter."""
        for field in fields(self):
            value = getattr(self, field.name)
            metadata = field.metadata
            if metadata['fixed']:
                continue
            support = metadata['support']
            if not isinstance(support, Support):
                raise TypeError
            yield value, support

    def parameters_name_value(self) -> Iterable[tuple[str, JaxComplexArray]]:
        """The name and value of each variable parameter."""
        for field in fields(self):
            name = field.name
            value = getattr(self, name)
            metadata = field.metadata
            if metadata['fixed']:
                continue
            yield name, value

    def parameters_name_value_support(self) -> Iterable[tuple[str, JaxComplexArray, Support]]:
        """The name, value, and support of each variable parameter."""
        for field in fields(self):
            name = field.name
            value = getattr(self, name)
            metadata = field.metadata
            if metadata['fixed']:
                continue
            support = metadata['support']
            if not isinstance(support, Support):
                raise TypeError
            yield name, value, support

    @classmethod
    def parameters_name_support(cls) -> Iterable[tuple[str, Support]]:
        """The name and support of each variable parameter."""
        for field in fields(cls):
            metadata = field.metadata
            if metadata['fixed']:
                continue
            support = metadata['support']
            if not isinstance(support, Support):
                raise TypeError
            yield field.name, support

    @property
    @abstractmethod
    def shape(self) -> Shape:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def domain_support(cls) -> Support:
        raise NotImplementedError



from __future__ import annotations

import numpy as np
import scipy.special
import scipy.stats as ss
from numpy.random import Generator
from tjax import NumpyComplexArray, NumpyRealArray, ShapeLike
from typing_extensions import override

from .shaped_distribution import ShapedDistribution

__all__ = ['ScipyDirichlet', 'ScipyGeneralizedDirichlet']


scipy_beta = scipy.special.beta


# pylint: disable=protected-access
mvd: type = ss._multivariate.dirichlet_frozen  # noqa: SLF001
class ScipyDirichletFixRVsAndPDF(mvd):
    """This class repairs dirichlet.

    See https://github.com/scipy/scipy/issues/6005 and https://github.com/scipy/scipy/issues/6006.
    """
    @override
    def rvs(self,
            size: ShapeLike | None = None,
            random_state: Generator | None = None) -> NumpyRealArray:
        if size is None:
            size = ()
        # This somehow fixes the behaviour of rvs.
        return super().rvs(size=size, random_state=random_state)  # pyright: ignore

    @override
    def pdf(self, x: NumpyRealArray) -> NumpyRealArray:
        if x.ndim == 2:  # noqa: PLR2004
            return super().pdf(x.T)
        return super().pdf(x)


class ScipyDirichlet(ShapedDistribution[ScipyDirichletFixRVsAndPDF]):
    """This class allows distributions having a non-empty shape."""
    @override
    def __init__(self, alpha: NumpyRealArray):
        shape = alpha[..., -1].shape
        rvs_shape = (alpha.shape[-1],)
        dtype = alpha.dtype
        objects = np.empty(shape, dtype=ScipyDirichletFixRVsAndPDF)
        for i in np.ndindex(shape):
            objects[i] = ScipyDirichletFixRVsAndPDF(alpha[i])
        super().__init__(shape, rvs_shape, dtype, objects)

    @override
    def pdf(self, x: NumpyComplexArray) -> NumpyRealArray:
        x = x.astype(np.float64)
        y = np.sum(x, axis=-1)
        if not np.all(np.isclose(y, np.ones(y.shape), atol=1e-5, rtol=0)):
            raise ValueError
        return super().pdf(x)


class ScipyGeneralizedDirichlet:
    @override
    def __init__(self, alpha: NumpyRealArray, beta: NumpyRealArray):
        super().__init__()
        self.alpha = alpha
        self.beta = beta

    def pdf(self, x: NumpyRealArray) -> NumpyRealArray:
        alpha_roll = np.roll(self.alpha, -1, axis=-1)
        alpha_roll[..., -1] = 0.0
        gamma = -np.diff(self.beta, append=1.0) - alpha_roll
        cs_x = np.cumsum(x, axis=-1)
        terms = x ** (self.alpha - 1.0) * (1.0 - cs_x) ** gamma / scipy_beta(self.alpha, self.beta)
        return np.prod(terms, axis=-1)

    def rvs(self, size: ShapeLike = (), random_state: Generator | None = None) -> NumpyComplexArray:
        sample_size: tuple[int, ...] = (((size,) if isinstance(size, int) else tuple(size))
                                        + self.alpha.shape)
        if random_state is None:
            random_state = np.random.default_rng()
        dimensions = self.alpha.shape[-1]
        beta_samples = random_state.beta(self.alpha, self.beta, size=sample_size)
        q = np.zeros(beta_samples.shape[:-1])
        for i in range(dimensions):
            beta_samples[..., i] *= 1 - q
            q += beta_samples[..., i]
        return beta_samples

    def entropy(self) -> NumpyRealArray:
        raise RuntimeError



from __future__ import annotations

from abc import abstractmethod

from jax.random import KeyArray
from tjax import JaxArray, Shape

from ..parametrization import Parametrization

__all__ = ['Samplable']


class Samplable(Parametrization):
    @abstractmethod
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxArray:
        raise NotImplementedError



from __future__ import annotations

from typing import Any

import jax
import jax.numpy as jnp
import numpy as np
from jax.random import KeyArray
from tjax import JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.multidimensional import Multidimensional
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import SymmetricMatrixSupport, VectorSupport, distribution_parameter

__all__ = ['MultivariateNormalNP', 'MultivariateNormalEP', 'MultivariateNormalVP']


def _broadcasted_outer(x: JaxRealArray) -> JaxRealArray:
    return jnp.einsum("...i,...j->...ij", x, x)


@dataclass
class MultivariateNormalNP(HasEntropyNP,
                           NaturalParametrization['MultivariateNormalEP', JaxRealArray],
                           Multidimensional):
    """The natural parametrization of the multivariate normal distribution.

    Args:
        mean_times_precision: E(x) / Var(x).
        negative_half_precision: -0.5 / Var(x).
    """
    mean_times_precision: JaxRealArray = distribution_parameter(VectorSupport())
    negative_half_precision: JaxRealArray = distribution_parameter(SymmetricMatrixSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_times_precision.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        eta = self.mean_times_precision
        h_inv = jnp.linalg.inv(self.negative_half_precision)
        a: JaxRealArray = jnp.einsum("...i,...ij,...j", eta, h_inv, eta)
        _, ld = jnp.linalg.slogdet(-self.negative_half_precision)
        return -0.25 * a - 0.5 * ld + 0.5 * self.dimensions() * jnp.log(np.pi)

    @override
    def to_exp(self) -> MultivariateNormalEP:
        h_inv = jnp.linalg.inv(self.negative_half_precision)
        h_inv_times_eta: JaxRealArray = jnp.einsum(
            "...ij,...j->...i", h_inv, self.mean_times_precision)
        mean = -0.5 * h_inv_times_eta
        second_moment = 0.25 * _broadcasted_outer(h_inv_times_eta) - 0.5 * h_inv
        return MultivariateNormalEP(mean, second_moment)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        return jnp.zeros(x.shape[:-1])

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> MultivariateNormalEP:
        return MultivariateNormalEP(x, _broadcasted_outer(x))

    @override
    def dimensions(self) -> int:
        return self.mean_times_precision.shape[-1]


@dataclass
class MultivariateNormalEP(HasEntropyEP[MultivariateNormalNP],
                           ExpectationParametrization[MultivariateNormalNP], Multidimensional,
                           Samplable):
    mean: JaxRealArray = distribution_parameter(VectorSupport())
    second_moment: JaxRealArray = distribution_parameter(SymmetricMatrixSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[MultivariateNormalNP]:
        return MultivariateNormalNP

    @override
    def to_nat(self) -> MultivariateNormalNP:
        precision = jnp.linalg.inv(self.variance())
        mean_times_precision = jnp.einsum("...ij,...j->...i", precision, self.mean)
        return MultivariateNormalNP(mean_times_precision, -0.5 * precision)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        return jnp.zeros(self.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_variance_parametrization().sample(key, shape)

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def variance(self) -> JaxRealArray:
        return self.second_moment - _broadcasted_outer(self.mean)

    def to_variance_parametrization(self) -> MultivariateNormalVP:
        return MultivariateNormalVP(self.mean, self.variance())


@dataclass
class MultivariateNormalVP(Samplable, Multidimensional):
    """The expectation parametrization of the multivariate normal distribution.

    Args:
        mean: E(x).
        variance: Var(x).
    """
    mean: JaxRealArray = distribution_parameter(VectorSupport())
    variance: JaxRealArray = distribution_parameter(SymmetricMatrixSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        return jax.random.multivariate_normal(key, self.mean, self.variance, shape)

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def to_exp(self) -> MultivariateNormalEP:
        second_moment = self.variance + _broadcasted_outer(self.mean)
        return MultivariateNormalEP(self.mean, second_moment)



from __future__ import annotations

from typing import Any

import jax.numpy as jnp
from tjax import Array, JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..mixins.transformed_parametrization import (TransformedExpectationParametrization,
                                                  TransformedNaturalParametrization)
from ..parameter import ScalarSupport, distribution_parameter
from .exponential import ExponentialEP, ExponentialNP

__all__ = ['RayleighNP', 'RayleighEP']


@dataclass
class RayleighNP(HasEntropyNP,
                 TransformedNaturalParametrization[ExponentialNP, ExponentialEP, 'RayleighEP',
                                                   JaxRealArray]):
    """The natural parametrization of the Rayleigh distribution.

    Args:
        eta: -1 / (2 * sigma^2).
    """
    eta: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.eta.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    @classmethod
    def base_distribution_cls(cls) -> type[ExponentialNP]:
        return ExponentialNP

    @override
    def base_distribution(self) -> ExponentialNP:
        return ExponentialNP(self.eta)

    @override
    @classmethod
    def create_expectation(cls, expectation_parametrization: ExponentialEP) -> RayleighEP:
        return RayleighEP(expectation_parametrization.mean)

    @override
    @classmethod
    def sample_to_base_sample(cls, x: Array, **fixed_parameters: Any
                              ) -> JaxRealArray:
        return jnp.square(x)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        return jnp.log(x) + jnp.log(2)


@dataclass
class RayleighEP(HasEntropyEP[RayleighNP],
                 TransformedExpectationParametrization[ExponentialEP, ExponentialNP, RayleighNP]):
    """The expectation parametrization of the Rayleigh distribution.

    Args:
        chi: 2 * sigma^2.
    """
    chi: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.chi.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[RayleighNP]:
        return RayleighNP

    @override
    @classmethod
    def base_distribution_cls(cls) -> type[ExponentialEP]:
        return ExponentialEP

    @override
    def base_distribution(self) -> ExponentialEP:
        return ExponentialEP(self.chi)

    @override
    def create_natural(self, natural_parametrization: ExponentialNP) -> RayleighNP:
        return RayleighNP(natural_parametrization.negative_rate)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        return 0.5 * jnp.log(self.chi * 0.5) + (1.5 * jnp.log(2.0) - 0.5 * jnp.euler_gamma)



from __future__ import annotations

from collections.abc import Iterable
from functools import reduce
from typing import TYPE_CHECKING, Any

import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from tjax import JaxComplexArray, JaxRealArray, NumpyComplexArray, NumpyRealArray
from typing_extensions import override

__all__: list[str] = []


def np_abs_square(x: NumpyComplexArray) -> NumpyRealArray:
    return np.square(x.real) + np.square(x.imag)  # pyright: ignore


def parameters_dot_product(x: NaturalParametrization[Any, Any], y: Any) -> JaxRealArray:
    def dotted_fields() -> Iterable[JaxRealArray]:
        for (x_value, x_support), (y_value, y_support) in zip(x.parameters_value_support(),
                                                              y.parameters_value_support(),
                                                              strict=True):
            axes = x_support.axes()
            assert y_support.axes() == axes
            yield _parameter_dot_product(x_value, y_value, axes)
    return reduce(jnp.add, dotted_fields())


iv_ratio = tfp.math.bessel_iv_ratio
log_ive = tfp.math.log_bessel_ive
betaln = tfp.math.lbeta


def vectorized_tril(m: NumpyRealArray, k: int = 0) -> NumpyRealArray:
    n, m_ = m.shape[-2:]
    indices = (..., *np.tril_indices(n, k, m_))
    values = m[indices]
    retval = np.zeros_like(m)
    retval[indices] = values
    return retval


def vectorized_triu(m: NumpyRealArray, k: int = 0) -> NumpyRealArray:
    n, m_ = m.shape[-2:]
    indices = (..., *np.triu_indices(n, k, m_))
    values = m[indices]
    retval = np.zeros_like(m)
    retval[indices] = values
    return retval


@override
def create_diagonal(m: NumpyRealArray) -> NumpyRealArray:
    """A vectorized version of diagonal.

    Args:
        m: Has shape (*k, n)
    Returns: Array with shape (*k, n, n) and the elements of m on the diagonals.
    """
    indices = (..., *np.diag_indices(m.shape[-1]))
    retval = np.zeros((*m.shape, m.shape[-1]), dtype=m.dtype)
    retval[indices] = m
    return retval


# Private functions --------------------------------------------------------------------------------
def _parameter_dot_product(x: JaxComplexArray, y: JaxComplexArray, n_axes: int) -> JaxRealArray:
    """Returns the real component of the dot product of the final n_axes axes of two arrays."""
    axes = tuple(range(-n_axes, 0))
    return jnp.sum(x * y, axis=axes).real


if TYPE_CHECKING:
    from .natural_parametrization import NaturalParametrization



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Generic, TypeVar, final

from tjax import JaxAbstractClass, JaxRealArray, abstract_jit, jit

from ..expectation_parametrization import ExpectationParametrization
from ..natural_parametrization import NaturalParametrization
from ..tools import parameters_dot_product

__all__ = ['HasEntropyEP', 'HasEntropyNP']


NP = TypeVar('NP', bound=NaturalParametrization[Any, Any])


class HasEntropyEP(ExpectationParametrization[NP],
                   JaxAbstractClass,
                   Generic[NP]):
    @abstract_jit
    @abstractmethod
    def expected_carrier_measure(self) -> JaxRealArray:
        """The expected carrier measure of the distribution.

        This is the missing term from the inner product between the observed distribution and the
        predicted distribution.  Often, it is just jnp.zeros(self.shape).
        """
        raise NotImplementedError

    @jit
    @final
    def cross_entropy(self, q: NP) -> JaxRealArray:
        """The cross entropy.

        Args:
            q: The natural parameters of the prediction.
        """
        return (-parameters_dot_product(q, self)
                + q.log_normalizer()
                - self.expected_carrier_measure())

    @jit
    @final
    def entropy(self) -> JaxRealArray:
        """The Shannon entropy.

        This can be quite slow since it depends on a conversion to natural parameters.
        """
        return self.cross_entropy(self.to_nat())


class HasEntropyNP(NaturalParametrization[HasEntropyEP[Any], Any]):
    @jit
    @final
    def entropy(self) -> JaxRealArray:
        """The Shannon entropy."""
        return self.to_exp().cross_entropy(self)
